from gama_config.generate_urdf import generate_urdf
from gama_config.gama_vessel import ArmidaleVesselConfig
from greenstream_config import Offsets


def test_generate_urdf():
    urdf = generate_urdf(
        config=ArmidaleVesselConfig(),
        ins_offset=Offsets(),
        mesh_path="package://some_package/meshes/some_mesh.stl",
    )

    assert """<mesh filename="package://some_package/meshes/some_mesh.stl"/>""" in urdf



import pytest
from pathlib import Path
from gama_config.gama_vessel import (
    read_vessel_config,
    WaveflyerVesselConfig,
    Mode,
    Variant,
    LogLevel,
    Network,
    EducatVesselConfig,
)
from gama_config.test.helpers import write_temp_file


def test_read_vessel_config_works():
    config_string = "\n".join(
        [
            "variant: educat",
            "ros_domain_id: 0",
            "namespace_vessel: vessel_1",
            "namespace_groundstation: groundstation",
            "mode: simulator",
            "network: host",
            "prod: false",
            "log_level: info",
            "ubiquity_user: ''",
            "ubiquity_pass: ''",
            "ubiquity_ip: ''",
            "cameras: null",
            "record: false",
            "components:",
            "  autopilot:",
            "    pid: [1, 2, 3]",
            "advanced_configuration:",
            "  first_launch_arg: value",
            "  second_launch_arg: value",
        ]
    )
    config_file = write_temp_file(config_string)

    config = read_vessel_config(config_file)

    assert config == EducatVesselConfig(
        ros_domain_id=0,
        namespace_vessel="vessel_1",
        namespace_groundstation="groundstation",
        mode=Mode.SIMULATOR,
        network=Network.HOST,
        prod=False,
        log_level=LogLevel.INFO,
        cameras=None,
        record=False,
        components={
            "autopilot": {"pid": [1, 2, 3]},
        },
        advanced_configuration={"first_launch_arg": "value", "second_launch_arg": "value"},
    )


def test_throws_if_file_not_found():
    with pytest.raises(FileNotFoundError, match="Could not find config file"):
        read_vessel_config(Path("does_not_exist.yaml"))


def test_throws_if_file_cannot_be_parsed():
    config_file = write_temp_file("")

    with pytest.raises(ValueError, match="Could not parse config file"):
        read_vessel_config(config_file)


def test_throws_if_mode_does_not_match_enum():
    config_string = "\n".join(
        [
            "cameras: null",
            "log_level: info",
            "mode: goblin",
            "network: host",
            "prod: false",
            "ubiquity_ip: ''",
            "ubiquity_pass: ''",
            "ubiquity_user: ''",
            "variant: educat",
        ]
    )
    config_file = write_temp_file(config_string)

    with pytest.raises(ValueError, match="Input should be 'simulator', 'hardware', 'stubs'"):
        read_vessel_config(config_file)


def test_throws_if_variant_does_not_match_enum():
    config_string = "\n".join(
        [
            "camera_overrides: null",
            "log_level: info",
            "mode: stubs",
            "network: host",
            "prod: false",
            "ubiquity_ip: ''",
            "ubiquity_pass: ''",
            "ubiquity_user: ''",
            "variant: killer-robot",
        ]
    )

    config_file = write_temp_file(config_string)

    with pytest.raises(
        ValueError,
        match=" Input tag 'killer-robot' found using 'variant' does not match any of the expected tags",
    ):
        read_vessel_config(config_file)


def test_throws_if_cameras_is_bad():
    config_string = "\n".join(
        [
            "cameras:",
            "- name: bow",
            "  order: 0",
            "  type: 89",
            "log_level: info",
            "mode: stubs",
            "network: host",
            "prod: false",
            "ubiquity_ip: ''",
            "ubiquity_pass: ''",
            "ubiquity_user: ''",
            "variant: educat",
        ]
    )
    config_file = write_temp_file(config_string)

    with pytest.raises(ValueError, match="Input should be a valid string"):
        read_vessel_config(config_file)


def test_throws_if_variant_specific_config_not_found():
    config_string = "\n".join(
        [
            "variant: waveflyer",
            "ros_domain_id: 0",
            "namespace_vessel: vessel_1",
            "namespace_groundstation: groundstation",
            "mode: simulator",
            "network: host",
            "prod: false",
            "log_level: info",
            "ubiquity_user: ''",
            "ubiquity_pass: ''",
            "ubiquity_ip: ''",
            "cameras: null",
            "record: false",
            "advanced_configuration:",
            "  random_launch_arg: value",
        ]
    )
    config_file = write_temp_file(config_string)

    config = read_vessel_config(config_file)

    assert config == WaveflyerVesselConfig(
        variant=Variant.WAVEFLYER,
        ros_domain_id=0,
        namespace_vessel="vessel_1",
        namespace_groundstation="groundstation",
        mode=Mode.SIMULATOR,
        network=Network.HOST,
        prod=False,
        log_level=LogLevel.INFO,
        cameras=None,
        record=False,
        components={},
        advanced_configuration={"random_launch_arg": "value"},
    )



from math import radians
from gama_config.gama_vessel import GamaVesselConfig, Variant
from greenstream_config import Offsets, get_cameras_urdf
from gr_urchin import URDF, Joint, Material, Link, xyz_rpy_to_matrix, Visual, Mesh, Geometry, Box


def generate_urdf(
    config: GamaVesselConfig,
    ins_offset: Offsets,
    mesh_path: str | None = None,
    waterline=0.0,  # meters between the waterline and the base_link
    radar_height=6.552,
    add_optical_frame: bool = True,
):
    file_path = f"/tmp/vessel_{config.variant.value}_{config.mode.value}.urdf"
    camera_links, camera_joints = get_cameras_urdf(config.cameras or [], add_optical_frame)

    urdf = URDF(
        name="origins",
        materials=[
            Material(name="grey", color=[0.75, 0.75, 0.75, 0.6]),
            Material(name="blue", color=[0, 0.12, 0.25, 0.9]),
        ],
        links=[
            Link(name="ins_link", inertial=None, visuals=None, collisions=None),
            Link(
                name="waterline",
                inertial=None,
                visuals=[
                    Visual(
                        name="waterline",
                        geometry=Geometry(box=Box([10.0, 10.0, 0.01])),
                        material=Material(name="blue"),
                    )
                ],
                collisions=None,
            ),
            Link(
                name="base_link",
                inertial=None,
                visuals=(
                    [
                        Visual(
                            name="visual",
                            geometry=Geometry(
                                mesh=Mesh(
                                    filename=mesh_path, combine=False, lazy_filename=mesh_path
                                )
                            ),
                            origin=xyz_rpy_to_matrix([0, 0, 0, radians(-90), 0, 0]),
                            material=Material(name="grey"),
                        ),
                    ]
                    if mesh_path
                    else None
                ),
                collisions=None,
            ),
            *camera_links,
        ],
        joints=[
            Joint(
                name="base_to_ins",
                parent="base_link",
                child="ins_link",
                joint_type="fixed",
                origin=xyz_rpy_to_matrix(
                    [
                        ins_offset.forward,
                        ins_offset.left,
                        ins_offset.up,
                        ins_offset.roll,
                        ins_offset.pitch,
                        ins_offset.yaw,
                    ]
                ),
            ),
            Joint(
                name="base_to_waterline",
                parent="base_link",
                child="waterline",
                joint_type="fixed",
                origin=xyz_rpy_to_matrix([0, 0, -waterline, 0, 0, 0]),
            ),
            *camera_joints,
        ],
    )
    # Add a radar
    if config.variant == Variant.ARMIDALE:
        urdf._links.append(
            Link(
                name="radar",
                inertial=None,
                visuals=[],
                collisions=None,
            )
        )
        urdf._joints.append(
            Joint(
                name="baselink_to_radar",
                parent="base_link",
                child="radar",
                joint_type="fixed",
                origin=xyz_rpy_to_matrix([0.0, 0.0, radar_height, 0.0, 0.0, radians(-1.5)]),
            )
        )

    urdf.save(file_path)

    # stringify urdf response for robot description
    with open(file_path) as infp:
        robot_description = infp.read()

    return robot_description



# IMPORTANT
# After changing this file, run `python3 -m gama_config.generate_schemas`
# To re-generate the json schemas

import os
import yaml
from pathlib import Path
from enum import Enum
from typing import Optional, Any, List, Union, Literal
from gama_config import LogLevel
from gama_config.helpers import write_config, read_config, find_gama_config, serialise
from greenstream_config.types import Camera
from pydantic import Field, BaseModel, ConfigDict, RootModel

GAMA_VESSEL_FILE_NAME = "gama_vessel.yml"
GAMA_VESSEL_SCHEMA_URL = (
    "https://greenroom-robotics.github.io/gama/schemas/gama_vessel.schema.json"
)


class Mode(str, Enum):
    SIMULATOR = "simulator"
    HARDWARE = "hardware"
    STUBS = "stubs"
    HITL_SIMULATOR = "hitl_simulator"

    def __str__(self):
        return self.value


class Network(str, Enum):
    SHARED = "shared"
    HOST = "host"

    def __str__(self):
        return self.value


class Variant(str, Enum):
    WHISKEY_BRAVO = "whiskey_bravo"
    EDUCAT = "educat"
    ORACLE_2_2 = "oracle_2_2"
    ORACLE_22 = "oracle_22"
    ARMIDALE = "armidale"
    WAVEFLYER = "waveflyer"
    DMAK = "dmak"
    MARS = "mars"

    def __str__(self):
        return self.value


class GamaVesselConfig(BaseModel):
    model_config = ConfigDict(
        json_encoders={
            Variant: lambda v: v.value,
            Mode: lambda v: v.value,
            LogLevel: lambda v: v.value,
            Network: lambda v: v.value,
        },
    )

    variant: Variant
    ros_domain_id: int = Field(
        default=0,
        description="ROS domain ID for the vessel. This only applies if 'simple_discovery' is true.",
    )
    namespace_vessel: str = "vessel_1"
    namespace_groundstation: str = "groundstation"
    mode: Mode = Mode.SIMULATOR
    network: Network = Network.HOST
    prod: bool = False
    log_level: LogLevel = LogLevel.INFO
    cameras: Optional[List[Camera]] = None
    record: bool = False
    advanced_configuration: Optional[dict[str, str]] = None
    components: Any = Field(default_factory=dict)
    log_directory: str = "~/greenroom/gama/logs"
    charts_directory: str = "~/greenroom/charts"
    simple_discovery: bool = Field(
        default=False,
        description="Use simple discovery. This should NOT be used in production due to performance issues.",
    )
    with_discovery_server: bool = Field(
        default=True, description="Run the discovery server. It will bind to 0.0.0.0:11811"
    )
    discovery_server_ip: str = Field(
        default="0.0.0.0",
        description="IP/host/interface of the discovery server. Assumes port of 11811",
    )
    own_ip: str = Field(
        default="0.0.0.0",
        description="IP/host/interface address of the primary network interface. This is where DDS traffic will route to.",
    )


class ArmidaleVesselConfig(GamaVesselConfig):
    variant: Literal[Variant.ARMIDALE] = Variant.ARMIDALE


class EducatVesselConfig(GamaVesselConfig):
    variant: Literal[Variant.EDUCAT] = Variant.EDUCAT


class Oracle22VesselConfig(GamaVesselConfig):
    variant: Literal[Variant.ORACLE_22] = Variant.ORACLE_22


class Oracle2_2VesselConfig(GamaVesselConfig):
    variant: Literal[Variant.ORACLE_2_2] = Variant.ORACLE_2_2


class WaveflyerVesselConfig(GamaVesselConfig):
    variant: Literal[Variant.WAVEFLYER] = Variant.WAVEFLYER


class WhiskeyBravoVesselConfig(GamaVesselConfig):
    variant: Literal[Variant.WHISKEY_BRAVO] = Variant.WHISKEY_BRAVO


class DMAKVesselConfig(GamaVesselConfig):
    variant: Literal[Variant.DMAK] = Variant.DMAK


class MarsVesselConfig(GamaVesselConfig):
    variant: Literal[Variant.MARS] = Variant.MARS


VariantVesselConfig = Union[
    ArmidaleVesselConfig,
    EducatVesselConfig,
    Oracle22VesselConfig,
    Oracle2_2VesselConfig,
    WaveflyerVesselConfig,
    WhiskeyBravoVesselConfig,
    DMAKVesselConfig,
    MarsVesselConfig,
]

DEFAULT_VARIANT_CONFIGS_MAP: dict[Variant, VariantVesselConfig] = {
    Variant.ARMIDALE: ArmidaleVesselConfig(),
    Variant.EDUCAT: EducatVesselConfig(),
    Variant.ORACLE_22: Oracle22VesselConfig(),
    Variant.ORACLE_2_2: Oracle2_2VesselConfig(),
    Variant.WAVEFLYER: WaveflyerVesselConfig(),
    Variant.WHISKEY_BRAVO: WhiskeyBravoVesselConfig(),
    Variant.DMAK: DMAKVesselConfig(),
    Variant.MARS: MarsVesselConfig(),
}


class VariantVesselConfigRoot(RootModel):
    root: VariantVesselConfig = Field(..., discriminator="variant")


def parse_vessel_config(config: Any) -> VariantVesselConfig:
    return VariantVesselConfigRoot(root=config).root


def get_vessel_config_path():
    return find_gama_config() / GAMA_VESSEL_FILE_NAME


def read_vessel_config(path: Optional[Path] = None) -> VariantVesselConfig:
    return read_config(path or get_vessel_config_path(), parse_vessel_config)


def read_vessel_config_env() -> VariantVesselConfig:
    config_str = os.environ.get("GAMA_VESSEL_CONFIG")
    if config_str is None:
        raise ValueError("GAMA_VESSEL_CONFIG environment variable not set")
    return parse_vessel_config(yaml.safe_load(config_str))


def write_vessel_config(config: GamaVesselConfig):
    return write_config(get_vessel_config_path(), config, GAMA_VESSEL_SCHEMA_URL)


def serialise_vessel_config(config: GamaVesselConfig):
    return serialise(config)



import yaml
import os
from typing import Any, Callable
from pathlib import Path
import json

type_parse = Callable[[Any], Any]


class YamlDumper(yaml.Dumper):
    """
    A YAML dumpler that show lists on the same line if they do not contain dicts or list
    """

    def represent_sequence(self, tag, sequence, flow_style=None):
        if isinstance(sequence, list) and all(
            [not isinstance(item, (dict, list)) for item in sequence]
        ):
            flow_style = True
        return super().represent_sequence(tag, sequence, flow_style)

    def represent_mapping(self, tag, mapping, flow_style=None):
        flow_style = False
        return super().represent_mapping(tag, mapping, flow_style)


def join_lines(*lines: str) -> str:
    return "\n".join(lines)


def find_gama_config() -> Path:
    """Returns the path to the .gama directory"""
    return Path.home().joinpath(".config/greenroom")


def write_config(path: Path, config: Any, schema_url: str):

    # Make the parent dir if it doesn't exist
    os.makedirs(path.parent, exist_ok=True)
    with open(path, "w") as stream:
        print(f"Writing: {path}")
        headers = f"# yaml-language-server: $schema={schema_url}"
        data = "\n".join(
            [
                headers,
                yaml.dump(json.loads(config.model_dump_json()), Dumper=YamlDumper, sort_keys=True),
            ]
        )
        stream.write(data)


def serialise(obj: Any) -> str:
    return yaml.dump(obj.model_dump_json(), default_flow_style=True, sort_keys=True)

    return "woop"


def read_config(path: Path, parse: type_parse):
    try:
        with open(path) as stream:
            return parse(yaml.safe_load(stream))
    except FileNotFoundError:
        raise FileNotFoundError(f"Could not find config file: {path}")
    except Exception as e:
        raise ValueError(f"Could not parse config file {path} - {e}")



# IMPORTANT
# After changing this file, run `python3 -m gama_config.generate_schemas`
# To re-generate the json schemas

import os
import yaml
from pathlib import Path
from pydantic import BaseModel, Field
from enum import Enum
from typing import Any, Optional
from gama_config import LogLevel
from gama_config.helpers import write_config, read_config, find_gama_config, serialise


GAMA_GS_FILE_NAME = "gama_gs.yml"
GAMA_GS_SCHEMA_URL = "https://greenroom-robotics.github.io/gama/schemas/gama_gs.schema.json"


class Variant(str, Enum):
    WHISKEY_BRAVO = "whiskey_bravo"
    EDUCAT = "educat"
    ORACLE_2_2 = "oracle_2_2"
    ORACLE_22 = "oracle_22"
    ARMIDALE = "armidale"
    WAVEFLYER = "waveflyer"
    DMAK = "dmak"
    MARS = "mars"

    def __str__(self):
        return self.value


class Mode(str, Enum):
    NONE = "none"
    XBOX = "xbox"
    XBOX_SERIES_X = "xbox_series_x"
    THRUSTMASTER = "thrustmaster"
    THRUSTMASTER_COMBO = "thrustmaster_combo"
    WARTHOG = "warthog"
    WARTHOG_COMBO = "warthog_combo"
    AERONAV = "aeronav"
    SINGLE_UNKNOWN = "single_unknown"
    DUAL_UNKNOWN = "dual_unknown"
    GLADIATOR = "gladiator"
    LOGITECH_EXTREME = "logitech_extreme"


class Network(str, Enum):
    SHARED = "shared"
    HOST = "host"


class GamaGsConfig(BaseModel):
    ros_domain_id: int = 0
    namespace_vessel: str = "vessel_1"
    namespace_groundstation: str = "groundstation"
    variant: Variant = Variant.ARMIDALE
    mode: Mode = Mode.NONE
    buttons: bool = False
    network: Network = Network.SHARED
    prod: bool = False
    log_level: LogLevel = LogLevel.INFO
    remote_cmd_override: bool = False
    simple_discovery: bool = False
    discovery_server_ip: str = Field(
        default="0.0.0.0",
        description="IP address of the discovery server. This should be the IP of the vessel.",
    )
    own_ip: str = Field(
        default="0.0.0.0",
        description="The IP address of the ground-stations's primary network interface. This is where DDS will bind to.",
    )


def parse_gs_config(config: dict[str, Any]) -> GamaGsConfig:
    # return from_dict(GamaGsConfig, config, config=Config(cast=[Mode, Network, LogLevel]))
    return GamaGsConfig(**config)


def get_gs_config_path():
    return find_gama_config() / GAMA_GS_FILE_NAME


def read_gs_config(path: Optional[Path] = None) -> GamaGsConfig:
    return read_config(path or get_gs_config_path(), parse_gs_config)


def read_gs_config_env() -> GamaGsConfig:
    gs_config_str = os.getenv("GAMA_GS_CONFIG")
    if gs_config_str is None:
        raise ValueError("GAMA_GS_CONFIG environment variable not set")
    return parse_gs_config(yaml.safe_load(gs_config_str))


def write_gs_config(config: GamaGsConfig):
    return write_config(get_gs_config_path(), config, GAMA_GS_SCHEMA_URL)


def serialise_gs_config(config: GamaGsConfig):
    return serialise(config)



import os
import json
from gama_config.gama_vessel import VariantVesselConfigRoot
from gama_config.gama_gs import GamaGsConfig
from pydantic.json_schema import GenerateJsonSchema


class CustomGenerateJsonSchema(GenerateJsonSchema):
    def generate(self, schema, mode="validation"):
        json_schema = super().generate(schema, mode=mode)
        json_schema["$schema"] = self.schema_dialect
        return json_schema


# TODO: fix schema generation for any of based json schema
def generate_schemas():
    """Generates the schemas for the config files"""

    SCHEMAS_PATH = os.path.join(os.path.dirname(__file__), "schemas")
    with open(os.path.join(SCHEMAS_PATH, "gama_vessel.schema.json"), "w") as f:
        main_model_schema = VariantVesselConfigRoot.model_json_schema()
        json.dump(main_model_schema, f, indent=2)
    with open(os.path.join(SCHEMAS_PATH, "gama_gs.schema.json"), "w") as f:
        main_model_schema = GamaGsConfig.model_json_schema()
        json.dump(main_model_schema, f, indent=2)


if __name__ == "__main__":
    print("Generating schemas...")
    generate_schemas()



from pathlib import Path


def write_temp_file(content: str) -> Path:
    """
    Write a temporary file with the given content and return the path to it.
    """
    import tempfile

    temp = tempfile.NamedTemporaryFile(delete=False)
    temp.write(content.encode())
    temp.close()
    return Path(temp.name)
